#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Argument parsing and script setup for the main glue.py script."""
from __future__ import annotations

import argparse
import os
import sys
from glob import glob
from typing import Callable, Optional

from polar2grid.core.script_utils import ExtendAction
from polar2grid.utils.dynamic_imports import get_reader_attr, get_writer_attr

# type aliases
ComponentParserFunc = Callable[[argparse.ArgumentParser], tuple]


def get_reader_parser_function(reader_name: str) -> Optional[ComponentParserFunc]:
    return get_reader_attr(reader_name, "add_reader_argument_groups")


def get_writer_parser_function(writer_name: str) -> Optional[ComponentParserFunc]:
    return get_writer_attr(writer_name, "add_writer_argument_groups")


READER_ALIASES = {
    "modis": "modis_l1b",
    "avhrr": "avhrr_l1b_aapp",
}
WRITER_ALIASES = {
    "scmi": "awips_tiled",
}


def get_default_output_filename(reader: str, writer: str, is_polar2grid: bool):
    """Get a default output filename based on what reader we are reading."""
    ofile_map = get_writer_attr(writer, "DEFAULT_OUTPUT_FILENAMES", {})
    pkg_name = "polar2grid" if is_polar2grid else "geo2grid"
    package_filenames = ofile_map[pkg_name]
    if reader not in package_filenames:
        reader = None
    return package_filenames.get(reader)


def _fsfiles_for_s3(input_filenames):
    """Convert S3 URLs to something Satpy can understand and use.

    Examples:
        Example S3 URLs (no caching):

        .. code-block:: bash

            polar2grid.sh ... -f s3://noaa-goes16/ABI-L1b-RadC/2019/001/17/*_G16_s20190011702186*

        Example S3 URLs using fsspec caching:

        .. code-block:: bash

            polar2grid.sh ... -f simplecache::s3://noaa-goes16/ABI-L1b-RadC/2019/001/17/*_G16_s20190011702186*

    """
    import fsspec
    from satpy.readers import FSFile

    kwargs = {"anon": True}
    if "simplecache::" in input_filenames[0]:
        kwargs = {"s3": kwargs}
    for open_file in fsspec.open_files(input_filenames, **kwargs):
        yield FSFile(open_file)


def _filenames_from_local(input_filenames):
    for fn in input_filenames:
        if os.path.isdir(fn):
            yield from glob(os.path.join(fn, "*"))
        else:
            yield fn


def get_input_files(input_filenames):
    """Convert directories to list of files."""
    if input_filenames and "s3://" in input_filenames[0]:
        yield from _fsfiles_for_s3(input_filenames)
    else:
        yield from _filenames_from_local(input_filenames)


def get_p2g_defaults_env_var():
    return bool(int(os.environ.setdefault("USE_POLAR2GRID_DEFAULTS", "1")))


class GlueArgumentParser:
    """Helper class for parsing and grouping command line arguments."""

    def __init__(self, argv: list[str], is_polar2grid: bool):
        self._is_polar2grid = is_polar2grid
        parser, known_args, reader_group, resampling_group, writer_group = _main_args(argv, self._is_polar2grid)
        # env used by some writers for default number of threads (ex. geotiff)
        # must be done before writer args are parsed
        os.environ["DASK_NUM_WORKERS"] = str(known_args.num_workers)

        reader_subgroups = _add_component_parser_args(parser, "readers", known_args.readers or [])
        writer_subgroups = _add_component_parser_args(parser, "writers", known_args.writers or [])
        self.argv = argv
        self._args = parser.parse_args(argv)
        _validate_reader_writer_args(parser, self._args, self._is_polar2grid)

        self._reader_args = _args_to_dict(self._args, reader_group._group_actions)
        self._reader_names = self._reader_args.pop("readers")
        self._separate_scene_init_load_args(reader_subgroups)
        self._resample_args = _args_to_dict(self._args, resampling_group._group_actions)
        self._writer_args = _args_to_dict(self._args, writer_group._group_actions)
        writer_specific_args = self._parse_one_writer_args(writer_subgroups)
        self._writer_args.update(writer_specific_args)

        if not self._args.filenames:
            parser.print_usage()
            parser.exit(1, "\nERROR: No data files provided (-f flag)\n")

    def _separate_scene_init_load_args(self, reader_subgroups) -> None:
        products = self._reader_args.pop("products") or []
        filenames = self._reader_args.pop("filenames") or []
        filenames = list(get_input_files(filenames))

        reader_specific_args, reader_specific_load_args = self._parse_reader_args(reader_subgroups)
        # argparse will combine "extended" arguments like `products` automatically
        # and products should only be provided to the load arguments, not reader creation
        for _reader_name, _reader_args in reader_specific_args.items():
            _reader_args.pop("products", None)

        # Parse provided files and search for files if provided directories
        self._scene_creation = {
            "filenames": filenames,
            "reader": self._reader_names[0],
            "reader_kwargs": reader_specific_args.get(self._reader_names[0], {}),
        }
        self._load_args = {
            "products": products,
        }
        self._load_args.update(reader_specific_load_args)

    def _parse_reader_args(self, reader_subgroups: list) -> tuple[dict, dict]:
        reader_args = {}
        load_args = {}
        for reader_name, (sgrp1, sgrp2) in zip(self._reader_names, reader_subgroups):
            if sgrp1 is None:
                continue
            rargs = _args_to_dict(self._args, sgrp1._group_actions)
            reader_args[reader_name] = rargs
            if sgrp2 is not None:
                load_args.update(_args_to_dict(self._args, sgrp2._group_actions))
        return reader_args, load_args

    def _parse_one_writer_args(self, writer_subgroups: list) -> dict:
        writer_names: list[str] = self._writer_args["writers"]
        writer_specific_args = {}
        for writer_name, (sgrp1, sgrp2) in zip(writer_names, writer_subgroups):
            wargs = _args_to_dict(self._args, sgrp1._group_actions)
            if sgrp2 is not None:
                wargs.update(_args_to_dict(self._args, sgrp2._group_actions))
            writer_specific_args[writer_name] = wargs
            # get default output filename
            if "filename" in wargs and wargs["filename"] is None:
                wargs["filename"] = get_default_output_filename(self._reader_names[0], writer_name, self._is_polar2grid)
        return writer_specific_args


def _main_args(argv, use_polar2grid_defaults):
    binary_name = "polar2grid" if use_polar2grid_defaults else "geo2grid"
    prog = os.getenv("PROG_NAME", sys.argv[0])
    # "usage: " will be printed at the top of this:
    usage = """
    %(prog)s -h
see available products:
    %(prog)s -r <reader> -w <writer> --list-products -f file1 [file2 ...]
basic processing:
    %(prog)s -r <reader> -w <writer> [options] -f file1 [file2 ...]
basic processing with limited products:
    %(prog)s -r <reader> -w <writer> [options] -p prod1 prod2 -f file1 [file2 ...]
"""
    parser = argparse.ArgumentParser(
        prog=prog,
        usage=usage,
        fromfile_prefix_chars="@",
        description="Load, composite, resample, and save datasets.",
    )
    _add_common_arguments(parser, binary_name)
    reader_group = add_scene_argument_groups(parser, is_polar2grid=use_polar2grid_defaults)[0]
    resampling_group = add_resample_argument_groups(parser, is_polar2grid=use_polar2grid_defaults)[0]
    writer_group = add_writer_argument_groups(parser, is_polar2grid=use_polar2grid_defaults)[0]
    argv_without_help = [x for x in argv if x not in ["-h", "--help"]]

    _retitle_optional_arguments(parser)
    args, _ = parser.parse_known_args(argv_without_help)
    return parser, args, reader_group, resampling_group, writer_group


def _add_common_arguments(parser: argparse.ArgumentParser, binary_name: str) -> None:
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="each occurrence increases verbosity 1 level through " "ERROR-WARNING-INFO-DEBUG (default INFO)",
    )
    parser.add_argument("-l", "--log", dest="log_fn", default=None, help="specify the log filename")
    parser.add_argument(
        "--progress",
        action="store_true",
        help="Show processing progress bar (Not recommended for logged output)",
    )
    parser.add_argument(
        "--create-profile",
        nargs="?",
        default=False,
        help=argparse.SUPPRESS,
        # help="Create an HTML document profiling the execution of the script "
        #      "using dask diagnostic tools.",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=os.getenv("DASK_NUM_WORKERS", 4),
        help="Specify number of worker threads to use (Default: 4)",
    )
    parser.add_argument(
        "--extra-config-path",
        action="append",
        help="Specify the base directory of additional Satpy configuration "
        "files. For example, to use custom enhancement YAML file named "
        "'generic.yaml' place it in a directory called 'enhancements' "
        "like '/path/to/my_configs/enhancements/generic.yaml' and then "
        "set this flag to '/path/to/my_configs'. For backwards compatibility, "
        "with older versions, this flag can also be given a single enhancement "
        "YAML configuration file.",
    )
    parser.add_argument(
        "--match-resolution",
        dest="preserve_resolution",
        action="store_false",
        help="When using the 'native' resampler for composites, don't save data "
        "at its native resolution, use the resolution used to create the "
        "composite.",
    )
    parser.add_argument(
        "--list-products",
        dest="list_products",
        action="store_true",
        help="List available {} products and exit".format(binary_name),
    )
    parser.add_argument(
        "--list-products-all",
        dest="list_products_all",
        action="store_true",
        help="List available {} products and custom/Satpy products and exit".format(binary_name),
    )


def _validate_reader_writer_args(parser, args, use_polar2grid_defaults):
    if args.readers is None:
        parser.print_usage()
        parser.exit(
            1,
            "\nERROR: Reader must be provided (-r flag).\n"
            "Supported readers:\n\t{}\n".format("\n\t".join(_supported_readers(use_polar2grid_defaults))),
        )
    elif len(args.readers) > 1:
        parser.print_usage()
        parser.exit(
            1,
            "\nMultiple readers is not currently supported. Got:\n\t" "{}\n".format("\n\t".join(args.readers)),
        )
        return -1
    if args.writers is None:
        parser.print_usage()
        parser.exit(
            1,
            "\nERROR: Writer must be provided (-w flag) with one or more writer.\n"
            "Supported writers:\n\t{}\n".format("\n\t".join(_supported_writers(use_polar2grid_defaults))),
        )


def _add_component_parser_args(
    parser: argparse.ArgumentParser, component_type: str, component_names: list[str]
) -> list:
    _get_args_func = get_writer_parser_function if component_type == "writers" else get_reader_parser_function
    subgroups = []
    for component_name in component_names:
        parser_func = _get_args_func(component_name)
        if parser_func is None:
            # two items in each tuple
            subgroups += [(None, None)]
            continue
        subgroups += [parser_func(parser)]
    return subgroups


def _args_to_dict(args, group_actions, exclude=None):
    if exclude is None:
        exclude = []
    return {
        ga.dest: getattr(args, ga.dest) for ga in group_actions if hasattr(args, ga.dest) and ga.dest not in exclude
    }


def _convert_reader_name(reader_name: str) -> str:
    return READER_ALIASES.get(reader_name, reader_name)


def _convert_writer_name(writer_name: str) -> str:
    return WRITER_ALIASES.get(writer_name, writer_name)


def _supported_readers(is_polar2grid: bool = False) -> list[str]:
    if is_polar2grid:
        readers = [
            "acspo",
            "amsr2_l1b",
            "amsr_l2_gaasp",
            "avhrr_l1b_aapp",
            "clavrx",
            "mersi2_l1b",
            "mirs",
            "modis_l1b",
            "modis_l2",
            "nucaps",
            "viirs_edr_active_fires",
            "viirs_edr_flood",
            "viirs_l1b",
            "viirs_sdr",
            "virr_l1b",
        ]
    else:
        readers = [
            "abi_l1b",
            "abi_l2_nc",
            "agri_fy4a_l1",
            "agri_fy4b_l1",
            "ahi_hrit",
            "ahi_hsd",
            "ami_l1b",
            "glm_l2",
        ]
    return sorted(readers)


def add_scene_argument_groups(parser, is_polar2grid=False):
    if is_polar2grid:
        filter_dn_products = 0.1
        filter_doc = " and is enabled by default."
    else:
        filter_dn_products = False
        filter_doc = ", but is disabled by default."
    readers = _supported_readers(is_polar2grid)

    group_1 = parser.add_argument_group(title="Reading")
    group_1.add_argument(
        "-r",
        "--reader",
        action="append",
        dest="readers",
        metavar="READER",
        type=_convert_reader_name,
        help="Name of reader used to read provided files. " "Supported readers: " + ", ".join(readers),
    )
    group_1.add_argument(
        "-f",
        "--filenames",
        nargs="+",
        default=[],
        help="Input files to read. For a long list of "
        "files, use '-f @my_files.txt' "
        "to provide a list of filenames from another "
        "file (one filename per line). Files can also "
        "be passed at the end of the command line by "
        "using two hyphens (--) to separate the list "
        "of files from the other arguments. "
        "arguments (ex. '%(prog)s ... -- /path/to/files*')",
    )
    group_1.add_argument("filenames", nargs="*", action="extend", help=argparse.SUPPRESS)
    # help="Alternative to '-f' flag. Use two hyphens (--) "
    #      "to separate these files from other command line "
    #      "arguments (ex. '%(prog)s ... -- /path/to/files*')")
    group_1.add_argument(
        "-p",
        "--products",
        nargs="+",
        default=None,
        action=ExtendAction,
        help="Names of products to create from input files",
    )
    group_1.add_argument(
        "--filter-day-products",
        nargs="?",
        type=float_or_false,
        default=filter_dn_products,
        metavar="fraction_of_day",
        help="Don't produce products that require "
        "daytime data when most of the image "
        "is nighttime (ex. reflectances). The "
        "list of products checked is currently limited "
        "to reflectance bands and true and false color "
        "composites. Default is 0.1 (at least 10%% "
        "day)" + filter_doc,
    )
    group_1.add_argument(
        "--filter-night-products",
        nargs="?",
        type=float_or_false,
        default=filter_dn_products,
        metavar="fraction_of_night",
        help="Don't produce products that require "
        "nighttime data when most of the image "
        "is daytime. The "
        "list of products checked is currently limited "
        "to temperature difference fog "
        "composites. Default is 0.1 (at least 10%% "
        "night)" + filter_doc,
    )
    group_1.add_argument(
        "--sza-threshold",
        type=float,
        default=100,
        help="When making filter decisions based on amount "
        "of day or amount of night, use this solar "
        "zenith angle value as the transition point. "
        "Less than this is day, greater than or equal to "
        "this value is night.",
    )
    group_1.add_argument(
        "--no-persist-geolocation",
        action="store_true",
        # help="After data is loaded, compute all geolocation arrays and hold "
        #      "them in memory. This should allow processing to perform faster "
        #      "in most cases at the cost of higher memory usage. This will "
        #      "have the biggest impact on readers whose lon/lat arrays are "
        #      "expensive to load (ex. modis_l1b). This currently only works "
        #      "on swath-based geolocation data and has no effect otherwise.",
        help=argparse.SUPPRESS,
    )
    return (group_1,)


def float_or_false(val):
    if isinstance(val, str) and val.lower() == "false":
        return False
    return float(val)


def _supported_writers(is_polar2grid: bool = False) -> list[str]:
    if is_polar2grid:
        writers = [
            "geotiff",
            "awips_tiled",
            "binary",
            "hdf5",
        ]
    else:
        writers = [
            "geotiff",
            "awips_tiled",
        ]
    return writers


def add_writer_argument_groups(parser, is_polar2grid=False):
    writers = _supported_writers(is_polar2grid)
    group_1 = parser.add_argument_group(title="Writing")
    group_1.add_argument(
        "-w",
        "--writer",
        action="append",
        dest="writers",
        type=_convert_writer_name,
        metavar="WRITER",
        help="Writer used to save datasets. " "Supported writers: " + ", ".join(writers),
        # help="Writers to save datasets with. Multiple writers "
        #      "can be provided by specifying '-w' multiple "
        #      "times (ex. '-w geotiff -w awips_tiled'). "
        #      "Supported writers: " + ", ".join(writers),
    )
    return (group_1,)


def add_resample_argument_groups(parser, is_polar2grid=None):
    if is_polar2grid is None:
        # if we are being loaded from documentation then this won't be
        # specified so we need to load it directly from the environment variable
        is_polar2grid = get_p2g_defaults_env_var()

    group_1 = parser.add_argument_group(title="Resampling")
    if is_polar2grid:
        DEBUG_EWA = bool(int(os.getenv("P2G_EWA_LEGACY", "0")))
        methods = ["ewa", "native", "nearest"]
        if DEBUG_EWA:
            methods.append("ewa_legacy")

        group_1.add_argument(
            "--method",
            dest="resampler",
            default=None,
            choices=methods,
            help="resampling algorithm to use (default: <sensor specific>)",
        )
        group_1.add_argument(
            "-g",
            "--grids",
            default=None,
            nargs="*",
            help="Area definition to resample to. Empty means "
            'no resampling (default: "wgs84_fit" for '
            "non-native resampling)",
        )

        # EWA options
        group_1.add_argument(
            "--weight-delta-max",
            default=argparse.SUPPRESS,
            type=float,
            help="Maximum distance in grid cells over which "
            'to distribute an input swath pixel (--method "ewa"). '
            'This is equivalent to the old "--fornav-D" flag. '
            "Default is 10.0.",
        )
        group_1.add_argument(
            "--weight-distance-max",
            default=argparse.SUPPRESS,
            type=float,
            help="Distance in grid cell units at which to "
            'apply a minimum weight. (--method "ewa"). '
            'This is equivalent to the old "--fornav-d" flag. '
            "Default is 1.0.",
        )
        group_1.add_argument(
            "--maximum-weight-mode",
            dest="maximum_weight_mode",
            default=argparse.SUPPRESS,
            action="store_true",
            help='Use maximum weight mode (--method "ewa"). Default is off.',
        )
        group_1.add_argument(
            "--rows-per-scan",
            dest="rows_per_scan",
            default=argparse.SUPPRESS,
            type=int,
            help="Number of data rows making up one instrument scan. "
            '(--method "ewa"). Defaults to value extracted from '
            "reader.",
        )
        group_1.add_argument(
            "--ewa-persist",
            action="store_true",
            help=argparse.SUPPRESS,
            # help="Pre-compute geolocation to determine what "
            # "chunks of data should be resampled "
            # '(--method "ewa"). This can reduce the '
            # "overall work required by the algorithm, "
            # "but only if a small amount of swath data "
            # "lies on the target area.",
        )
    else:
        group_1.add_argument(
            "--method",
            dest="resampler",
            default=None,
            choices=["native", "nearest"],
            help="resampling algorithm to use (default: native)",
        )
        group_1.add_argument(
            "-g",
            "--grids",
            default=None,
            nargs="*",
            help='Area definition to resample to. Empty means no resampling (default: "MAX")',
        )

    # shared options
    group_1.add_argument(
        "--grid-coverage",
        default=0.1,
        type=float,
        help="Fraction of target grid that must contain " "data to continue processing product.",
    )
    group_1.add_argument(
        "--cache-dir",
        help="Directory to store resampling intermediate "
        "results between executions. Not used with native "
        "resampling or resampling of ungridded or swath data.",
    )
    group_1.add_argument(
        "--grid-configs",
        nargs="+",
        default=tuple(),
        help="Specify additional grid configuration files. "
        "(.conf for legacy CSV grids, .yaml for "
        "SatPy-style areas)",
    )
    group_1.add_argument(
        "--ll-bbox",
        nargs=4,
        type=float,
        metavar=("lon_min", "lat_min", "lon_max", "lat_max"),
        help=argparse.SUPPRESS
        if is_polar2grid
        else "Crop data to region specified by lon/lat "
        "bounds (lon_min lat_min lon_max lat_max). "
        "Coordinates must be valid in the source data "
        "projection. Can only be used with gridded "
        "input data.",
    )
    group_1.add_argument(
        "--antimeridian-mode",
        default="modify_crs",
        choices=("modify_extents", "modify_crs", "global_extents"),
        help="Behavior when dynamic grids are converted to 'frozen' grids and "
        "data crosses the anti-meridian. Defaults to 'modify_crs' where "
        "the prime meridian is shifted 180 degrees to make the result one "
        "contiguous coordinate space. 'modify_extents' will attempt to "
        "surround the data but will often cause artifacts over the "
        "antimeridian. 'global_extents' will force the X extents to -180 "
        "and 180 to create one large grid. This currently only affects "
        "lon/lat projections.",
    )

    # nearest neighbor resampling
    group_1.add_argument(
        "--radius-of-influence",
        default=argparse.SUPPRESS,
        type=float,
        help="Specify radius to search for valid input "
        'pixels for nearest neighbor resampling (--method "nearest"). '
        "Value is in geocentric meters regardless of input or output projection. "
        "By default this will be estimated based on input and output projection and pixel size.",
    )
    return tuple([group_1])


def _retitle_optional_arguments(parser):
    """Hack to make the optional arguments say what we want."""
    opt_args = [x for x in parser._action_groups if x.title == "optional arguments"]
    if len(opt_args) == 1:
        opt_args = opt_args[0]
        opt_args.title = "Global Options"



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2019 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The VIIRS EDR Active Fires reader operates on CSPP NetCDF I-Band (AFIMG) Resolution or
M-Band Resolution (AFMOD) Environmental Data Record files.

Files supported usually have the following naming schemes:

   AFIMG_j01_d20221006_t2101052_e2102297_b25304_c20221006214545032016_cspp_dev.nc and/or,
   AFMOD_npp_d20221006_t2017005_e2018247_b56701_c20221006205259096916_cspp_dev.nc

For more information about the this CSPP product, please
visit the CSPP LEO website: `https://cimss.ssec.wisc.edu/cspp/`.

This reader's default resampling algorithm ``--method`` is
``nearest`` for Nearest Neighbor resampling. The frontend can
be specified with the ``polar2grid.sh`` command using the
``viirs_edr_active_fires`` frontend name. The VIIRS Active Fire
frontend provides the following products:

+---------------------------+-----------------------------------------------------+
| **Product Name**          | **Description**                                     |
+===========================+=====================================================+
| confidence_cat            | Fire Confidence Category   (AFIMG Resolution Only)  |
+---------------------------+-----------------------------------------------------+
| T4                        | I-Band 4 Temperature       (AFIMG Resolution Only)  |
+---------------------------+-----------------------------------------------------+
| power                     | Fire Radiative Power                                |
+---------------------------+-----------------------------------------------------+
| confidence_pct            | Fire Confidence Percentage (AFMOD Resolution Only)  |
+---------------------------+-----------------------------------------------------+
| T13                       | M-Band 13 Temperature      (AFMOD Resolution Only)  |
+---------------------------+-----------------------------------------------------+
"""
from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from satpy import DataQuery

from ._base import ReaderProxyBase

DEFAULT_DATASETS = ["T4", "T13", "confidence_cat", "confidence_pct", "power"]


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_DATASETS

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return DEFAULT_DATASETS

    @property
    def _aliases(self) -> dict[str, DataQuery]:
        return {}


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="VIIRS EDR Active Fires Reader")
    return group, None



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Base class for all filters."""

import logging

from satpy import Scene
from xarray import DataArray

logger = logging.getLogger(__name__)


class BaseFilter:
    """Base class for filtering products that don't match certain conditions.

    This class uses a series of metadata comparisons to check if a product
    should be checked for filtering. If any of the product metadata criteria
    match then filtering checks are continued. Otherwise, the product is
    ignored. If the criteria is not provided or is ``None`` or is empty then
    no products will be checked. If the criteria is passed as ``True`` then
    all products will be checked.

    By default, no extra filtering is performed and only the metadata criteria
    is used. This means that if no criteria is provided, this class will
    remove all provided DataArrays.

    """

    FILTER_MSG = "Unloading '{}' due to filtering."

    def __init__(self, product_filter_criteria: dict = None):
        """Initialize thresholds and default search criteria."""
        self._filter_criteria = product_filter_criteria

    def _matches_criteria(self, data_arr: DataArray):
        attrs = data_arr.attrs
        if self._filter_criteria is True:
            # check all products
            return True
        if not self._filter_criteria:
            # if no criteria was provided then no products will be checked
            return False

        for filter_key, filter_list in self._filter_criteria.items():
            if attrs.get(filter_key) in filter_list:
                return True
        return False

    def _filter_data_array(self, data_arr: DataArray, _cache: dict):
        """Check if this DataArray should be removed.

        Returns:
            True if it meets the condition and does not have enough
            day data and should therefore be removed. False otherwise
            meaning it should be kept.

        """
        if not self._matches_criteria(data_arr):
            return False
        # Subclasses should implement further logic here
        return True

    def filter_scene(self, scene: Scene):
        """Create a new Scene with filtered DataArrays removed."""
        _cache = {}
        remaining_ids = []
        filtered_ids = []
        for data_id in self._iter_scene_coarsest_to_finest_area(scene):
            logger.debug("Analyzing '{}' for filtering...".format(data_id))
            if not self._filter_data_array(scene[data_id], _cache):
                remaining_ids.append(data_id)
            else:
                logger.debug(self.FILTER_MSG.format(data_id))
                filtered_ids.append(data_id)
        if not remaining_ids:
            return None
        new_scn = scene.copy(remaining_ids)
        new_scn._wishlist = scene.wishlist.copy() - set(filtered_ids)
        return new_scn

    @staticmethod
    def _iter_scene_coarsest_to_finest_area(scene: Scene):
        by_area_list = list(scene.iter_by_area())
        by_area_list = sorted(by_area_list, key=lambda x: x[0].shape[0])
        for _, data_arrays in by_area_list:
            yield from data_arrays



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2014 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
# Written by David Hoese    November 2014
# University of Wisconsin-Madison
# Space Science and Engineering Center
# 1225 West Dayton Street
# Madison, WI  53706
# david.hoese@ssec.wisc.edu
"""Script for installing the polar2grid package.

See the documentation site for more information:

http://www.ssec.wisc.edu/software/polar2grid/

"""
from glob import glob

from setuptools import find_packages, setup

version = "3.0.1"


def readme():
    with open("README.rst", "r") as f:
        return f.read()


classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: Implementation :: CPython",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: MacOS :: MacOS X",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Atmospheric Science",
    "Topic :: Scientific/Engineering :: GIS",
]

extras_require = {
    # Other:
    "utils": ["matplotlib"],
    "docs": ["sphinx", "rst2pdf", "sphinx-argparse", "sphinxcontrib-apidoc", "pytest"],
    "tests": ["pytest"],
    "coastlines": ["pycoast", "pydecorate"],
}
extras_require["all"] = list(set([x for y in extras_require.values() for x in y]))

entry_points = {
    "console_scripts": [
        "polar2grid=polar2grid.__main__:p2g_main",
        "geo2grid=polar2grid.__main__:g2g_main",
    ],
}

setup(
    name="polar2grid",
    version=version,
    author="David Hoese, SSEC",
    author_email="david.hoese@ssec.wisc.edu",
    license="GPLv3",
    description="Library and scripts to remap satellite data to a grid",
    long_description=readme(),
    classifiers=classifiers,
    keywords="",
    url="https://www.ssec.wisc.edu/software/polar2grid/",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "polar2grid": [
            "grids/*.conf",
        ]
    },
    # The location of where these are installed are important for the swbundle and glue scripts!
    # Look at env.sh, glue.py, and glue_legacy.py for where these are pointed to.
    data_files=[
        ("etc/polar2grid/colormaps", glob("etc/colormaps/*")),
        ("etc/polar2grid/composites", glob("etc/composites/*")),
        ("etc/polar2grid/enhancements", glob("etc/enhancements/*")),
        ("etc/polar2grid/readers", glob("etc/readers/*")),
        ("etc/polar2grid/writers", glob("etc/writers/*")),
        ("etc/polar2grid", ["etc/pyspectral.yaml"]),
        ("etc/polar2grid", ["etc/resampling.yaml"]),
    ],
    zip_safe=True,
    install_requires=[
        "setuptools",  # reading configuration files
        "satpy",
        "rasterio",
        "netCDF4",
        "h5py",
    ],
    python_requires=">=3.9",
    extras_require=extras_require,
    entry_points=entry_points,
    scripts=[file for file in glob("swbundle/*.sh") if file not in ["swbundle/env.sh", "swbundle/polar2grid_env.sh"]],
)



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Connect various satpy components together to go from satellite data to output imagery format."""

from __future__ import annotations

import contextlib
import logging
import os
import shutil
import sys
import tempfile
from collections.abc import Iterable
from datetime import datetime
from typing import Optional, Union

import dask
import satpy
from dask.diagnostics import ProgressBar
from pyresample import SwathDefinition
from satpy import DataID, Scene
from satpy.writers import compute_writer_results

from polar2grid._glue_argparser import GlueArgumentParser, get_p2g_defaults_env_var
from polar2grid.core.script_utils import (
    create_exc_handler,
    rename_log_file,
    setup_logging,
)
from polar2grid.filters import filter_scene
from polar2grid.readers._base import ReaderProxyBase
from polar2grid.resample import resample_scene
from polar2grid.utils.config import add_polar2grid_config_paths
from polar2grid.utils.dynamic_imports import get_reader_attr
from polar2grid.utils.legacy_compat import get_sensor_alias

LOG = logging.getLogger(__name__)

_PLATFORM_ALIASES = {
    "suomi-npp": "npp",
    "snpp": "npp",
    "n20": "noaa20",
    "n21": "noaa21",
    "n22": "noaa22",
    "n23": "noaa23",
    "noaa-18": "noaa18",
    "noaa-19": "noaa19",
    "noaa-20": "noaa20",
    "noaa-21": "noaa21",
    "noaa-22": "noaa22",
    "noaa-23": "noaa23",
    "jpss-1": "noaa20",
    "jpss-2": "noaa21",
    "jpss-3": "noaa22",
    "jpss-4": "noaa23",
    "j1": "noaa20",
    "j2": "noaa21",
    "j3": "noaa22",
    "j4": "noaa23",
    "fy-3b": "fy3b",
    "fy-3c": "fy3c",
    "fy-3d": "fy3d",
    "eos-aqua": "aqua",
    "eos-terra": "terra",
    "aqua": "aqua",
    "terra": "terra",
    "gcom-w1": "gcom-w1",
    "metop-a": "metopa",
    "metop-b": "metopb",
    "metop-c": "metopc",
}


def _get_platform_name_alias(satpy_platform_name):
    return _PLATFORM_ALIASES.get(satpy_platform_name.lower(), satpy_platform_name)


def _overwrite_platform_name_with_aliases(scn):
    """Change 'platform_name' for every DataArray to Polar2Grid expectations."""
    for data_arr in scn:
        if "platform_name" not in data_arr.attrs:
            continue
        pname = _get_platform_name_alias(data_arr.attrs["platform_name"])
        data_arr.attrs["platform_name"] = pname


def _overwrite_sensor_with_aliases(scn):
    """Change 'sensor' for every DataArray to Polar2Grid expectations."""
    for data_arr in scn:
        if "sensor" not in data_arr.attrs:
            continue
        pname = get_sensor_alias(data_arr.attrs["sensor"])
        data_arr.attrs["sensor"] = pname


def _write_scene(
    scn: Scene, writers: list[str], writer_args: dict[str, dict], data_ids: list[DataID], to_save: Optional[list] = None
):
    if to_save is None:
        to_save = []
    if not data_ids:
        # no datasets to save
        return to_save

    _assign_default_native_area_id(scn, data_ids)
    for writer_name in writers:
        wargs = writer_args[writer_name]
        _write_scene_with_writer(scn, writer_name, data_ids, wargs, to_save)
    return to_save


def _assign_default_native_area_id(scn: Scene, data_ids: list[DataID]) -> None:
    for data_id in data_ids:
        area_def = scn[data_id].attrs.get("area")
        if area_def is None or hasattr(area_def, "area_id"):
            continue
        scn[data_id].attrs["area"].area_id = "native"


def _write_scene_with_writer(scn: Scene, writer_name: str, data_ids: list[DataID], wargs: dict, to_save: list) -> None:
    res = scn.save_datasets(writer=writer_name, compute=False, datasets=data_ids, **wargs)
    if res and isinstance(res[0], (tuple, list)):
        # list of (dask-array, file-obj) tuples
        to_save.extend(zip(*res))
    else:
        # list of delayed objects
        to_save.extend(res)


def _print_list_products(reader_info, is_polar2grid: bool, p2g_only: bool):
    available_p2g_names, available_custom_names, available_satpy_names = reader_info.get_available_products()
    available_satpy_names = ["*" + _sname for _sname in available_satpy_names]
    available_custom_names = ["*" + _sname for _sname in available_custom_names]
    project_name = "Polar2Grid" if is_polar2grid else "Geo2Grid"

    print("### Custom User Products")
    print("\n".join(sorted(available_custom_names)) if available_custom_names else "<None>")
    print()

    if not p2g_only:
        print("### Non-standard Satpy Products")
        print("\n".join(sorted(available_satpy_names)) if available_satpy_names else "<None>")
        print()

    print(f"### Standard Available {project_name} Products")
    print("\n".join(sorted(available_p2g_names)) if available_p2g_names else "<None>")


def _create_scene(scene_creation: dict) -> Optional[Scene]:
    try:
        scn = Scene(**scene_creation)
    except ValueError as e:
        LOG.error("{} | Enable debug message (-vvv) or see log file for details.".format(str(e)))
        LOG.debug("Further error information: ", exc_info=True)
        return
    except OSError:
        LOG.error("Could not open files. Enable debug message (-vvv) or see log file for details.")
        LOG.debug("Further error information: ", exc_info=True)
        return
    return scn


def _resample_scene_to_grids(
    scn: Scene,
    reader_names: list[str],
    resample_args: dict,
    filter_kwargs: dict,
    preserve_resolution: bool,
    use_polar2grid_defaults: bool,
) -> list[tuple]:
    ll_bbox = resample_args.pop("ll_bbox")
    if ll_bbox:
        scn = scn.crop(ll_bbox=ll_bbox)

    scn = filter_scene(
        scn,
        reader_names,
        **filter_kwargs,
    )
    if scn is None:
        LOG.info("No remaining products after filtering.")
        return []

    areas_to_resample = resample_args.pop("grids")
    antimeridian_mode = resample_args.pop("antimeridian_mode")
    if "ewa_persist" in resample_args:
        resample_args["persist"] = resample_args.pop("ewa_persist")
    scenes_to_save = resample_scene(
        scn,
        areas_to_resample,
        antimeridian_mode=antimeridian_mode,
        preserve_resolution=preserve_resolution,
        is_polar2grid=use_polar2grid_defaults,
        **resample_args,
    )
    return scenes_to_save


def _save_scenes(scenes_to_save: list[tuple], reader_info, writer_args) -> list:
    to_save = []
    for scene_to_save, products_to_save in scenes_to_save:
        _overwrite_platform_name_with_aliases(scene_to_save)
        _overwrite_sensor_with_aliases(scene_to_save)
        reader_info.apply_p2g_name_to_scene(scene_to_save)
        to_save = _write_scene(
            scene_to_save,
            writer_args["writers"],
            writer_args,
            products_to_save,
            to_save=to_save,
        )
    return to_save


def _get_glue_name(args):
    reader_name = "NONE" if args.readers is None else args.readers[0]
    writer_names = "-".join(args.writers or [])
    return f"{reader_name}_{writer_names}"


@contextlib.contextmanager
def _create_profile_html_if(create_profile: Union[False, None, str], project_name: str, glue_name: str):
    from dask.diagnostics import CacheProfiler, Profiler, ResourceProfiler, visualize

    if create_profile is False:
        yield
        return
    if create_profile is None:
        profile_filename = "{project_name}_{glue_name}_{start_time:%Y%m%d_%H%M%S}.html"
    else:
        profile_filename = create_profile

    start_time = datetime.now()
    with CacheProfiler() as cprof, ResourceProfiler() as rprof, Profiler() as prof:
        yield
    end_time = datetime.now()

    profile_filename = profile_filename.format(
        project_name=project_name,
        glue_name=glue_name,
        start_time=start_time,
        end_time=end_time,
    )
    profile_filename = os.path.abspath(profile_filename)
    visualize([prof, rprof, cprof], file_path=profile_filename, show=False)
    print(f"Profile HTML: file://{profile_filename}")


def main(argv=sys.argv[1:]):
    ret = -1
    try:
        processor = _GlueProcessor(argv)
    except FileNotFoundError:
        return ret

    try:
        ret = processor()
    finally:
        processor.cleanup()
    return ret


class _GlueProcessor:
    """Helper class to make calling processing steps easier."""

    def __init__(self, argv):
        add_polar2grid_config_paths()
        self.is_polar2grid = get_p2g_defaults_env_var()
        self.arg_parser = GlueArgumentParser(argv, self.is_polar2grid)
        self.glue_name = _get_glue_name(self.arg_parser._args)
        self.rename_log = _prepare_initial_logging(self.arg_parser, self.glue_name)
        self.tmp_config_paths = []
        self._handle_extra_config_paths(self.arg_parser._args)
        self._clean = False

    def _handle_extra_config_paths(self, args):
        if not args.extra_config_path:
            return
        _check_valid_config_paths(args.extra_config_path)

        new_config_paths = []
        # Preserve user's specified order to handle inheritance/overrides
        for extra_config_path in args.extra_config_path:
            if os.path.isdir(extra_config_path):
                new_config_paths.append(extra_config_path)
                continue

            tmp_config_path = _create_tmp_enhancement_config_dir(extra_config_path)
            new_config_paths.append(tmp_config_path)
            self.tmp_config_paths.append(tmp_config_path)
        _add_extra_config_paths(new_config_paths)
        LOG.debug(f"Satpy config path is: {satpy.config.get('config_path')}")

    def cleanup(self):
        self._clean = True
        for tmp_config_path in self.tmp_config_paths:
            LOG.debug(f"Deleting temporary config directory: {tmp_config_path}")
            shutil.rmtree(tmp_config_path, ignore_errors=True)

    def __call__(self):
        # Set up dask and the number of workers
        common_args = self.arg_parser._args
        if common_args.num_workers:
            dask.config.set(num_workers=common_args.num_workers)

        with _create_profile_html_if(
            common_args.create_profile,
            "polar2grid" if self.is_polar2grid else "geo2grid",
            self.glue_name,
        ):
            return self._run_processing()

    def _run_processing(self):
        LOG.info("Sorting and reading input files...")
        arg_parser = self.arg_parser
        preferred_chunk_size = get_reader_attr(arg_parser._scene_creation["reader"], "PREFERRED_CHUNK_SIZE", 1024)
        _set_preferred_chunk_size(preferred_chunk_size)
        scn = _create_scene(arg_parser._scene_creation)
        if scn is None:
            return -1

        if self.rename_log:
            stime = getattr(scn, "start_time", scn.attrs.get("start_time"))
            rename_log_file(self.glue_name + stime.strftime("_%Y%m%d_%H%M%S.log"))

        # Load the actual data arrays and metadata (lazy loaded as dask arrays)
        LOG.info("Loading product metadata from files...")
        load_args = arg_parser._load_args.copy()
        user_products = load_args.pop("products")
        reader_info = ReaderProxyBase.from_reader_name(arg_parser._scene_creation["reader"], scn, user_products)
        if arg_parser._args.list_products or arg_parser._args.list_products_all:
            _print_list_products(reader_info, self.is_polar2grid, not arg_parser._args.list_products_all)
            return 0

        products = reader_info.get_satpy_products_to_load()
        persist_geolocation = not arg_parser._reader_args.pop("no_persist_geolocation", False)
        if not products:
            return -1
        scn.load(products, **load_args, generate=False)
        if persist_geolocation:
            scn = _persist_swath_definition_in_scene(scn)
        scn.generate_possible_composites(True)

        reader_args = arg_parser._reader_args
        filter_kwargs = {
            "sza_threshold": reader_args["sza_threshold"],
            "day_fraction": reader_args["filter_day_products"],
            "night_fraction": reader_args["filter_night_products"],
        }
        scenes_to_save = _resample_scene_to_grids(
            scn,
            arg_parser._reader_names,
            arg_parser._resample_args,
            filter_kwargs,
            arg_parser._args.preserve_resolution,
            self.is_polar2grid,
        )
        to_save = _save_scenes(scenes_to_save, reader_info, arg_parser._writer_args)

        if arg_parser._args.progress:
            pbar = ProgressBar()
            pbar.register()

        LOG.info("Computing products and saving data to writers...")
        if not to_save:
            LOG.warning(
                "No product files produced given available valid data and "
                "resampling settings. This can happen if the writer "
                "detects that no valid output will be written or the "
                "input data does not overlap with the target grid."
            )
        compute_writer_results(to_save)
        LOG.info("SUCCESS")
        return 0


def _prepare_initial_logging(arg_parser, glue_name: str) -> bool:
    global LOG
    LOG = logging.getLogger(glue_name)

    # Prepare logging
    args = arg_parser._args
    rename_log = False
    if args.log_fn is None:
        rename_log = True
        args.log_fn = glue_name + "_fail.log"
    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    setup_logging(console_level=levels[min(3, args.verbosity)], log_filename=args.log_fn)
    logging.getLogger("rasterio").setLevel(levels[min(1, args.verbosity)])
    logging.getLogger("fsspec").setLevel(levels[min(2, args.verbosity)])
    logging.getLogger("s3fs").setLevel(levels[min(2, args.verbosity)])
    logging.getLogger("aiobotocore").setLevel(levels[min(2, args.verbosity)])
    logging.getLogger("botocore").setLevel(levels[min(2, args.verbosity)])
    sys.excepthook = create_exc_handler(LOG.name)
    if levels[min(3, args.verbosity)] > logging.DEBUG:
        import warnings

        warnings.filterwarnings("ignore")
    LOG.debug("Starting script with arguments: %s", " ".join(arg_parser.argv))
    return rename_log


def _add_extra_config_paths(extra_paths: list[str]):
    config_path = satpy.config.get("config_path")
    LOG.info(f"Adding additional configuration paths: {extra_paths}")
    satpy.config.set(config_path=config_path + extra_paths)


def _check_valid_config_paths(extra_config_paths: Iterable):
    single_file_configs = [config_path for config_path in extra_config_paths if os.path.isfile(config_path)]
    config_paths = [config_path for config_path in extra_config_paths if os.path.isdir(config_path)]
    invalid_paths = set(extra_config_paths) - (set(single_file_configs) | set(config_paths))
    if invalid_paths:
        str_paths = "\n\t".join(sorted(invalid_paths))
        msg = f"Specified extra config paths don't exist:\n\t{str_paths}"
        LOG.error(msg)
        raise FileNotFoundError(msg)


def _create_tmp_enhancement_config_dir(enh_yaml_file: str) -> str:
    config_dir = tempfile.mkdtemp(prefix="p2g_tmp_config")
    enh_dir = os.path.join(config_dir, "enhancements")
    LOG.debug(f"Creating temporary config directory for enhancement file {enh_yaml_file!r}: {enh_dir}")
    os.makedirs(enh_dir)

    new_enh_file = os.path.join(enh_dir, "generic.yaml")
    shutil.copy(enh_yaml_file, new_enh_file)
    return config_dir


def _set_preferred_chunk_size(preferred_chunk_size: int) -> None:
    pcs_in_mb = (preferred_chunk_size * preferred_chunk_size) * 8 // (1024 * 1024)
    if "PYTROLL_CHUNK_SIZE" not in os.environ:
        LOG.debug(f"Setting preferred chunk size to {preferred_chunk_size} pixels or {pcs_in_mb:d}MiB")
        satpy.CHUNK_SIZE = preferred_chunk_size
        os.environ["PYTROLL_CHUNK_SIZE"] = f"{preferred_chunk_size:d}"
        dask.config.set({"array.chunk-size": f"{pcs_in_mb:d}MiB"})
    else:
        LOG.debug(f"Using environment variable chunk size: {os.environ['PYTROLL_CHUNK_SIZE']}")


def _persist_swath_definition_in_scene(scn: Scene) -> None:
    to_persist_swath_defs = _swaths_to_persist(scn)
    if not to_persist_swath_defs:
        return scn

    to_update_data_arrays, to_persist_lonlats = zip(*to_persist_swath_defs.values())
    LOG.info("Loading swath geolocation into memory...")
    persisted_lonlats = dask.persist(*to_persist_lonlats)
    persisted_swath_defs = [SwathDefinition(plons, plats) for plons, plats in persisted_lonlats]
    new_scn = scn.copy()
    for arrays_to_update, persisted_swath_def in zip(to_update_data_arrays, persisted_swath_defs):
        for array_to_update in arrays_to_update:
            array_to_update.attrs["area"] = persisted_swath_def
            new_scn._datasets[array_to_update.attrs["_satpy_id"]] = array_to_update
    LOG.debug(f"{len(to_persist_swath_defs)} unique swath definitions persisted")
    return new_scn


def _swaths_to_persist(scn: Scene) -> dict:
    to_persist_swath_defs = {}
    for data_arr in scn.values():
        swath_def = data_arr.attrs.get("area")
        if not isinstance(swath_def, SwathDefinition):
            continue
        this_swath_data_array_copies, _ = to_persist_swath_defs.setdefault(
            swath_def, ([], (swath_def.lons, swath_def.lats))
        )
        this_swath_data_array_copies.append(data_arr.copy())
    return to_persist_swath_defs


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Convert legacy grids.conf files to grids.yaml format."""

from __future__ import annotations

import logging
import os
import sys
import warnings

import yaml
from pyproj import CRS, Proj

from polar2grid.grids.manager import read_grids_config

logger = logging.getLogger(__name__)


def _conf_to_yaml_dict(grids_filename: str) -> str:
    overall_yaml_dict = {}
    grids_information = read_grids_config(grids_filename, convert_coords=False)
    for grid_name, grid_info in grids_information.items():
        try:
            area_dict = grid_info_to_yaml_dict(grid_info)
        except ValueError:
            continue

        overall_yaml_dict[grid_name] = area_dict
    return overall_yaml_dict


def grid_info_to_yaml_dict(grid_info: dict) -> dict:
    area_dict = {}
    area_dict["description"] = ""

    crs = grid_info["crs"] if "crs" in grid_info else CRS.from_proj4(grid_info["proj4_str"])
    proj_dict = crs_to_proj_dict(crs)
    area_dict["projection"] = proj_dict

    _add_shape(grid_info, area_dict)
    dx, dy = _add_resolution(grid_info, area_dict)
    _add_origin(grid_info, area_dict, crs, dx, dy)
    return area_dict


def crs_to_proj_dict(crs: CRS) -> dict:
    warnings.filterwarnings("ignore", module="pyproj", category=UserWarning)
    if crs.to_epsg() is not None:
        proj_dict = {"EPSG": crs.to_epsg()}
    else:
        proj_dict = crs.to_dict()
    _remove_unnecessary_proj_params(proj_dict)
    return proj_dict


def _remove_unnecessary_proj_params(proj_dict: dict) -> None:
    for param_name, exp_value in [("x_0", 0), ("y_0", 0), ("k_0", 1)]:
        if proj_dict.get(param_name) == exp_value:
            del proj_dict[param_name]


def _add_shape(grid_info: dict, area_dict: dict) -> None:
    width = grid_info["grid_width"]
    height = grid_info["grid_height"]
    if width is not None and height is not None:
        area_dict["shape"] = {"height": height, "width": width}


def _add_resolution(grid_info: dict, area_dict: dict) -> tuple[float, float]:
    dx = abs(grid_info["pixel_size_x"])
    dy = abs(grid_info["pixel_size_y"])
    if dx is not None and dy is not None:
        area_dict["resolution"] = {"dy": dy, "dx": dx}
    return dx, dy


def _add_origin(grid_info: dict, area_dict: dict, crs: CRS, dx: float, dy: float) -> None:
    ox = grid_info["grid_origin_x"]
    oy = grid_info["grid_origin_y"]
    if ox is not None and dx is None:
        logger.error("Can't convert grid with origin but no pixel resolution")
        raise ValueError("Can't convert grid with origin but no pixel resolution")
    if ox is not None and oy is not None:
        convert_to_meters = not crs.is_geographic and grid_info["grid_origin_units"] == "degrees"
        ox_m, oy_m = Proj(crs)(ox, oy) if convert_to_meters else (ox, oy)
        # convert center-pixel coordinates to outer edge extent
        ox_m = ox_m - dx / 2.0
        oy_m = oy_m + dy / 2.0
        ox, oy = Proj(crs)(ox_m, oy_m, inverse=True) if convert_to_meters else (ox_m, oy_m)
        area_dict["upper_left_extent"] = {"x": ox, "y": oy, "units": grid_info["grid_origin_units"]}


def ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
    """Dump the data to YAML in ordered fashion."""

    class OrderedDumper(Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items(), flow_style=False)

    OrderedDumper.add_representer(dict, _dict_representer)
    return yaml.dump(data, stream, OrderedDumper, **kwds)


def get_parser():
    from argparse import ArgumentParser

    prog = os.getenv("PROG_NAME", sys.argv[0])
    parser = ArgumentParser(
        prog=prog,
        description="Convert legacy grids.conf format to Pyresample YAML format.",
        usage="""
To write to a file:
    %(prog)s input_file.conf > output_file.yaml
""",
    )
    parser.add_argument("grids_filename", help="Input grids.conf-style file to convert to YAML.")
    return parser


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = get_parser()
    args = parser.parse_args(argv)
    logging.basicConfig(level=logging.INFO)
    yaml_dict = _conf_to_yaml_dict(args.grids_filename)
    yml_str = ordered_dump(yaml_dict, default_flow_style=None)
    print(yml_str)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2018 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The AHI HSD Reader operates on standard files from the Japan
Meteorological Agency (JMA) Himawari-8 and Himawari-9 Advanced
Himawari Imager (AHI) instrument. The AHI Himawari Standard Data
(HSD) reader works off of the input filenames to determine if a
file is supported by Geo2Grid. Files usually have the following
naming scheme:

    HS_H08_20181022_0300_B09_FLDK_R20_S1010.DAT

These are the mission compliant radiance file naming conventions
used by JMA. The AHI HSD reader supports all instrument spectral bands,
identified in Geo2Grid as the products shown in the table below. The
AHI HSD reader can be provided to the main geo2grid.sh script
using the ``-r`` option and the reader name ``ahi_hsd``.

The list of supported products includes true and natural color imagery.
These are created by means of a python based atmospheric Rayleigh
scattering correction algorithm that is executed as part of the |project| AHI
HSD reader, along with sharpening to the highest spatial resolution. For
more information on the creation of RGBs, please see the
:ref:`RGB section <getting_started_rgb>`.

+---------------------------+-----------------------------------------------------+
| **Product Name**          | **Description**                                     |
+===========================+=====================================================+
| B01                       | Channel 1 (0.47um) Reflectance Band                 |
+---------------------------+-----------------------------------------------------+
| B02                       | Channel 2 (0.51um) Reflectance Band                 |
+---------------------------+-----------------------------------------------------+
| B03                       | Channel 3 (0.64um) Reflectance Band                 |
+---------------------------+-----------------------------------------------------+
| B04                       | Channel 4 (0.86um) Reflectance Band                 |
+---------------------------+-----------------------------------------------------+
| B05                       | Channel 5 (1.6um) Reflectance Band                  |
+---------------------------+-----------------------------------------------------+
| B06                       | Channel 6 (2.3um) Reflectance Band                  |
+---------------------------+-----------------------------------------------------+
| B07                       | Channel 7 (3.9um) Brightness Temperature Band       |
+---------------------------+-----------------------------------------------------+
| B08                       | Channel 8 (6.2um) Brightness Temperature Band       |
+---------------------------+-----------------------------------------------------+
| B09                       | Channel 9 (6.9um) Brightness Temperature Band       |
+---------------------------+-----------------------------------------------------+
| B10                       | Channel 10 (7.3um) Brightness Temperature Band      |
+---------------------------+-----------------------------------------------------+
| B11                       | Channel 11 (8.6um) Brightness Temperature Band      |
+---------------------------+-----------------------------------------------------+
| B12                       | Channel 12 (9.6um) Brightness Temperature Band      |
+---------------------------+-----------------------------------------------------+
| B13                       | Channel 13 (10.4um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| B14                       | Channel 14 (11.2um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| B15                       | Channel 15 (12.4um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| B16                       | Channel 16 (13.3um) Brightness Temperature Band     |
+---------------------------+-----------------------------------------------------+
| true_color                | Ratio sharpened rayleigh corrected true color       |
+---------------------------+-----------------------------------------------------+
| natural_color             | Ratio sharpened rayleigh corrected natural color    |
+---------------------------+-----------------------------------------------------+
| airmass                   | Air mass RGB                                        |
+---------------------------+-----------------------------------------------------+
| ash                       | Ash RGB                                             |
+---------------------------+-----------------------------------------------------+
| dust                      | Dust RGB                                            |
+---------------------------+-----------------------------------------------------+
| fog                       | Fog RGB                                             |
+---------------------------+-----------------------------------------------------+
| night_microphysics        | Night Microphysics RGB                              |
+---------------------------+-----------------------------------------------------+

"""

from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from ..core.script_utils import BooleanOptionalAction
from ._base import ReaderProxyBase

PREFERRED_CHUNK_SIZE: int = 2200  # one segment

READER_PRODUCTS = ["B{:02d}".format(x) for x in range(1, 17)]
COMPOSITE_PRODUCTS = [
    "true_color",
    "natural_color",
    "airmass",
    "ash",
    "dust",
    "fog",
    "night_microphysics",
]
DEFAULT_PRODUCTS = READER_PRODUCTS + COMPOSITE_PRODUCTS[:2]


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_geo2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return READER_PRODUCTS + COMPOSITE_PRODUCTS[:2]

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return READER_PRODUCTS + COMPOSITE_PRODUCTS


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="AHI HSD Reader")
    group.add_argument("--mask-space", action=BooleanOptionalAction, default=True, help="Mask space pixels.")
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2020 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
#     Written by David Hoese    March 2020
#     University of Wisconsin-Madison
#     Space Science and Engineering Center
#     1225 West Dayton Street
#     Madison, WI  53706
#     david.hoese@ssec.wisc.edu
"""The AWIPS Tiled writer is used to create AWIPS compatible tiled NetCDF4 files.

The Advanced Weather Interactive Processing System (AWIPS) is a
program used by the United States National Weather Service (NWS) and
others to display and analyze data relevant to meteorology.
Sectorized Cloud and Moisture Imagery (SCMI) is a NetCDF4
format accepted by AWIPS to store one image broken up
into one or more "tiles". This format has been used to support additional
products over time and so this writer is now called "awips_tiled" to refer
to the generic usse of these files. Once AWIPS is configured for specific products
the AWIPS Tiled writer can be used to provide compatible products to the
system. The files created by this writer are compatible with AWIPS II.

The writer takes remapped image data and creates an
AWIPS-compatible NetCDF4 file. The writer and the AWIPS client may
need to be configured to make things appear the way the user wants in
the AWIPS client. The writer can only produce files for datasets mapped
to areas with specific projections:

    - Lambert Conformal Conic (`+proj=lcc`)
    - Geostationary (`+proj=geos`)
    - Mercator (`+proj=merc`)
    - Polar Stereographic (`+proj=stere`)

This is a limitation of the AWIPS client and not of the writer.

Numbered versus Lettered Grids
------------------------------

By default the writer will save tiles by number starting with '1'
representing the upper-left image tile. Tile numbers then increase
along the column and then on to the next row.

By specifying ``--letters`` on the command line, tiles can be designated with a
letter. Lettered grids or sectors are preconfigured in the writer
configuration file (``awips_tiled.yaml`` in SatPy). The lettered tile locations are
static and will not change with the data being written to them. Each lettered
tile is split in to a certain number of subtiles (`--letter-subtiles`),
default 2 rows by 2 columns. Lettered tiles are meant to make it easier for
receiving AWIPS clients/stations to filter what tiles they receive; saving
time, bandwidth, and space.

Tiles (numbered or lettered) not containing any valid data are not
created.

 .. warning::

     The writer does not default to using any grid. Therefore, it is recommended to specify
     one or more grids for remapping by using the `-g` flag.

For more detailed information on templates and other options for this writer
see the Satpy documentation :mod:`here <satpy.writers.awips_tiled>`.

"""
import logging

from polar2grid.core.script_utils import BooleanOptionalAction
from polar2grid.utils.legacy_compat import convert_p2g_pattern_to_satpy

LOG = logging.getLogger(__name__)

DEFAULT_OUTPUT_FILENAMES = {
    "polar2grid": {
        None: "{source_name}_AII_{platform_name}_{sensor!l}_{p2g_name}"
        "_{sector_id}_{tile_id}_{start_time:%Y%m%d_%H%M}.nc",
    },
    "geo2grid": {
        None: None,  # default to template filename
    },
}


def add_writer_argument_groups(parser, group=None):
    import argparse

    if group is None:
        group = parser.add_argument_group(title="AWIPS Tiled Writer")
    # group_1.add_argument('--file-pattern', default=DEFAULT_OUTPUT_PATTERN,
    #                      help="Custom file pattern to save dataset to")
    group.add_argument("--compress", action="store_true", help="zlib compress each netcdf file")
    group.add_argument("--fix-awips", action="store_true", help=argparse.SUPPRESS)
    # help="modify NetCDF output to work with the old/broken AWIPS NetCDF library")
    group.add_argument(
        "--output-filename",
        dest="filename",
        type=convert_p2g_pattern_to_satpy,
        help="custom file pattern to save dataset to",
    )
    group.add_argument(
        "--use-end-time",
        action="store_true",
        help="use end_time metadata inplace of start_time (useful for multi-day composites)",
    )
    group.add_argument(
        "--use-sector-reference",
        action="store_true",
        help="use the lettered sector location as reference "
        "and shift data to match tile pixel locations. "
        "Useful when tiles will be updated in future "
        "executions. By default the sector tiles are "
        "shifted to match the data location. Maximum "
        "shift is 0.5 pixels.",
    )
    # Saving specific keyword arguments
    # group_2 = parser.add_argument_group(title='Writer Save')
    group.add_argument(
        "--tiles",
        dest="tile_count",
        nargs=2,
        type=int,
        default=[1, 1],
        help="Number of tiles to produce in Y (rows) and X (cols) direction respectively",
    )
    group.add_argument(
        "--tile-size",
        dest="tile_size",
        nargs=2,
        type=int,
        default=None,
        help="Specify how many pixels are in each tile (overrides '--tiles')",
    )
    group.add_argument(
        "--letters",
        dest="lettered_grid",
        action="store_true",
        help="Create tiles from a static letter-based grid based on the product projection",
    )
    group.add_argument(
        "--letter-subtiles",
        nargs=2,
        type=int,
        default=(2, 2),
        help="Specify number of subtiles in each lettered tile: 'row col'",
    )
    group.add_argument(
        "--source-name",
        default="SSEC",
        help="specify processing source name used in attributes and filename",
    )
    group.add_argument(
        "--sector-id",
        help="specify name for sector/region used in attributes and filename (example 'LCC')",
    )
    group.add_argument(
        "--template",
        default="polar",
        help="specify name for pre-configured template used to determine output file structure and formatting.",
    )
    group.add_argument(
        "--environment-prefix",
        default="DR",
        help="Force value for templates that support an 'environment_prefix' in the filename (default 'OR').",
    )
    group.add_argument(
        "--check-categories",
        default=True,
        action=BooleanOptionalAction,
        help="Specify whether category/flag products should be checked for empty tiles. "
        "By default (True), tiles consisting entirely of 'missing' pixels will not "
        "be written. For category products this may not be desired as the missing "
        "or invalid value is something users are interested in.",
    )
    return group, None



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Utilities for dynamically imported and accessing components."""

import importlib
from typing import Any


def _get_component_attr(component_type: str, component_name: str, attr_name: str, default: Any = None) -> Any:
    try:
        comp_mod = importlib.import_module(f"polar2grid.{component_type}.{component_name}")
    except ModuleNotFoundError:
        return default
    return getattr(comp_mod, attr_name, default)


def get_reader_attr(reader_name: str, attr_name: str, default: Any = None) -> Any:
    return _get_component_attr("readers", reader_name, attr_name, default=default)


def get_writer_attr(writer_name: str, attr_name: str, default: Any = None) -> Any:
    return _get_component_attr("writers", writer_name, attr_name, default=default)



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Helper script for generating debug AWIPS Tiled files for verifying AWIPS client behavior."""
import contextlib
import os
import sys
import tempfile
from datetime import datetime

import dask.array as da
import numpy as np
import satpy
import xarray as xr
import yaml
from pyresample import create_area_def
from satpy import Scene

from polar2grid.utils.config import add_polar2grid_config_paths

SCALE_FACTOR = 0.25
ADD_OFFSET = 2000.0


def main():
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Create single AWIPS tile example file for debugging")
    parser.add_argument(
        "physical_element_prefix", help="Name of the product in AWIPS. No spaces or special characters allowed."
    )
    parser.add_argument(
        "--units", default="1", help="CF-compatible units. Will be converted if necessary to AWIPS-compatible units."
    )
    parser.add_argument(
        "--idtype", default="uint16", help="Numpy data type of the data created and then stored in the NetCDF file."
    )
    parser.add_argument(
        "--odtype", default="int16", help="Numpy data type of the data created and then stored in the NetCDF file."
    )
    parser.add_argument(
        "--ounsigned",
        action="store_true",
        help="Whether or not to include the special 'Unsigned' flag in the NetCDF file. No effect if odtype "
        "isn't signed.",
    )
    args = parser.parse_args()

    add_polar2grid_config_paths()

    start_time = datetime.utcnow()
    idtype_str = args.idtype
    idtype = getattr(np, idtype_str)
    odtype_str = args.odtype
    set_unsigned = args.ounsigned
    set_unsigned_str = "U" if set_unsigned else ""
    units = args.units
    physical_element = (
        args.physical_element_prefix.replace(" ", "-") + f"-{units}-{idtype_str}-{set_unsigned_str}{odtype_str}"
    )

    data = _gradient_data(dtype=idtype)
    area = create_area_def(
        "fakelcc",
        {"proj": "lcc", "lon_0": -95.0, "lat_0": 25, "lat_1": 25},
        shape=data.shape,
        resolution=(10000.0, 10000.0),
        center=(0, 0),
    )
    data_arr = xr.DataArray(
        data,
        attrs={
            "platform_name": "P2G-DEBUG",
            "sensor": "P2G-INST",
            "name": physical_element,
            "start_time": start_time,
            "units": units,
            "area": area,
            "valid_range": _valid_range_for_idtype(idtype),
        },
        dims=("y", "x"),
    )
    scn = Scene()
    scn[physical_element] = data_arr

    os.environ["ORGANIZATION"] = "P2G_DEBUG_ORG"
    with add_fake_awips_template(physical_element, odtype_str, set_unsigned):
        scn.save_datasets(
            writer="awips_tiled",
            sector_id="LCC",
            tile_count=(4, 4),
            source_name="SSEC",
        )


@contextlib.contextmanager
def add_fake_awips_template(product_name, odtype_str, set_unsigned):
    with tempfile.TemporaryDirectory(prefix="p2g_awips_debug_") as tdir:
        writers_dir = os.path.join(tdir, "writers")
        os.makedirs(writers_dir)
        yaml_filename = os.path.join(writers_dir, "awips_tiled.yaml")
        with open(yaml_filename, "w") as yaml_file:
            var_config = {
                "name": product_name,
                "var_name": "data",
                "attributes": {
                    "physical_element": {"value": "{name}"},
                    "units": {},
                },
                "encoding": {
                    "dtype": odtype_str,
                    # "scale_factor": SCALE_FACTOR,
                    # "add_offset": ADD_OFFSET,
                    # "_FillValue": 10,
                },
            }
            if set_unsigned:
                var_config["encoding"]["_Unsigned"] = "true"
            yaml.dump(
                {
                    "templates": {
                        "polar": {
                            "variables": {
                                "my_fake_var": var_config,
                            },
                        },
                    },
                },
                yaml_file,
            )

        with satpy.config.set(config_path=[tdir] + satpy.config.get("config_path")):
            yield


def _valid_range_for_idtype(dtype):
    min_val = np.iinfo(dtype).min
    max_val = np.iinfo(dtype).max
    return [min_val, max_val]


def _gradient_data(dtype):
    num_rows = 300
    num_cols = 100
    min_val, max_val = _valid_range_for_idtype(dtype)
    unscaled_min = min_val * SCALE_FACTOR + ADD_OFFSET
    unscaled_max = max_val * SCALE_FACTOR + ADD_OFFSET
    data = np.repeat(np.linspace(unscaled_min, unscaled_max, num_cols)[np.newaxis, :], num_rows, axis=0)
    half_row = num_rows // 2
    half_col = num_cols // 2
    data[half_row - 5 : half_row + 5, half_col - 5 : half_col + 5] = np.nan
    return da.from_array(data)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Helpers for setting up the Polar2Grid environment and configuration."""
import importlib.metadata as impm
import importlib.resources as impr
import os
import sys

import satpy


def get_polar2grid_etc():
    p2g_pkg_location = impr.files("polar2grid")
    if _is_editable_installation():
        return str(p2g_pkg_location.parent / "etc")
    return os.path.join(sys.prefix, "etc", "polar2grid")


def _is_editable_installation():
    for installed_file in impm.files("polar2grid"):
        str_fn = str(installed_file)
        if "__editable__" in str_fn or "pyproject.toml" in str_fn:
            return True
    return False


def get_polar2grid_home():
    p2g_home = os.environ.get("POLAR2GRID_HOME")
    if p2g_home is None:
        # assume development/editable install
        import polar2grid

        os.environ["POLAR2GRID_HOME"] = os.path.join(os.path.dirname(polar2grid.__file__), "..", "swbundle")
    return os.environ["POLAR2GRID_HOME"]


def add_polar2grid_config_paths():
    config_path = satpy.config.get("config_path")
    p2g_etc = get_polar2grid_etc()
    if p2g_etc not in config_path:
        satpy.config.set(config_path=config_path + [p2g_etc])



"""Special TOC directive to filter out specific enries based on configured options.

See https://stackoverflow.com/questions/15001888/conditional-toctree-in-sphinx
for more information.

"""
import re

from docutils.parsers.rst import directives
from sphinx.directives.other import TocTree

FILTER_OPTION_SPEC = {
    "excludebuilder": directives.unchanged,
}
FILTER_OPTION_SPEC.update(TocTree.option_spec)


def setup(app):
    app.add_config_value("toc_filter_exclude", [], "html")
    app.add_directive("toctree-filt", TocTreeFilt)
    return {"version": "1.0.0"}


class TocTreeFilt(TocTree):
    """Directive to allow filtering of TOC entries based on configured options.

    Directive to notify Sphinx about the hierarchical structure of the docs,
    and to include a table-of-contents like tree in the current document. This
    version filters the entries based on a list of prefixes. We simply filter
    the content of the directive and call the super's version of run. The
    list of exclusions is stored in the **toc_filter_exclusion** list. Any
    table of content entry prefixed by one of these strings will be excluded.
    If `toc_filter_exclusion=['secret','draft']` then all toc entries of the
    form `:secret:ultra-api` or `:draft:new-features` will be excuded from
    the final table of contents. Entries without a prefix are always included.
    """

    option_spec = FILTER_OPTION_SPEC
    hasPat = re.compile(r"^\s*:(.+):(.+)$")

    # Remove any entries in the content that we dont want and strip
    # out any filter prefixes that we want but obviously don't want the
    # prefix to mess up the file name.
    def filter_entries(self, entries):
        if self.env.app.builder.name in self.options.get("excludebuilder", ""):
            return []
        excl = self.state.document.settings.env.config.toc_filter_exclude
        filtered = []
        for e in entries:
            m = self.hasPat.match(e)
            if m is not None:
                if not m.groups()[0] in excl:
                    filtered.append(m.groups()[1])
            else:
                filtered.append(e)
        return filtered

    def run(self):
        # Remove all TOC entries that should not be on display
        self.content = self.filter_entries(self.content)
        return super().run()



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Tests for base/shared logic."""

import importlib
import os

import pytest

PKG_ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", ".."))
READER_BASE = os.path.join(PKG_ROOT, "readers")
ALL_READERS = [os.path.splitext(reader_fn)[0] for reader_fn in os.listdir(READER_BASE) if not reader_fn.startswith("_")]


@pytest.mark.parametrize("reader_name", ALL_READERS)
def test_reader_imports(reader_name):
    mod = importlib.import_module(f"polar2grid.readers.{reader_name}")
    assert hasattr(mod, "ReaderProxy")


@pytest.mark.parametrize("reader_name", ALL_READERS)
def test_reader_parser_arguments(reader_name):
    import argparse

    mod = importlib.import_module(f"polar2grid.readers.{reader_name}")
    assert hasattr(mod, "add_reader_argument_groups")

    parser = argparse.ArgumentParser()
    groups = mod.add_reader_argument_groups(parser)
    assert len(groups) == 2
